<?php
class string_analytics {
	// 分詞的來源字串
	private $sourceString;
	// 初步分析的結果
	private $simpleResult;
	// 分詞的結果
	private $finallyResult;
	// 字典檔
	private $ditc;
	// 文字解析方向
	private $gran_type;
	// 詞彙最大長度
	private $max_words;
	// 結論最多數量
	private $max_tags;

	function __construct() {
		$this->sourceString = '';
		$this->simpleResult = array();
		$this->n_gramResult = array();
		$this->finallyResult = array();
		$this->ditc = array();
	}

	public function get_finally() {
		$return = array();
		$i = $this->max_tags;
		foreach( $this->finallyResult as $key => $times ) {
			$return[] = $key;
			$i--;
			if( $i <= 0 ) {
				break;
			}
		}
		return $return;
	}

	public function SetOption($option) {
		$this->gran_type = $option['ngram_type'];
		$this->max_words = $option['max_words'];
		$this->max_tags = $option['max_tags'];
	}

	public function SetDict($ditc) {
		foreach( $ditc as $key => $value ) {
			if( isset($this->ditc[$key]) ) {
				$this->ditc[$key] += $value;
			} else {
				$this->ditc[$key] = $value;
			}
		}
	}

	public function SetSource($source) {
		$this->sourceString = $source;
		$this->htmltag_overwrite();
		$this->StringClear();
	}

	public function StartAnalysis() {
		$this->simpleString();
		if( $this->gran_type == 'f' ) {
			$this->forward_n_gram();
		} else {
			$this->reverse_n_gram();
		}
		$this->makefinally();
	}

	private function htmltag_overwrite() {
		preg_match_all('/<h[1-6][^>]*>([^<]*)<\/h[1-6]\s*>/si', $this->sourceString, $match);
		if( count($match[1]) > 0 ) {
			$this->sourceString .= ' ' . implode(' ', $match[1]);
		}
		while( preg_match_all('/<[b|i|em|strong][^>]*>([^<]*)<\/[b|i|em|strong]\s*>/si', $this->sourceString, $match) > 0 ) {
			$this->sourceString = str_replace($match[0], $match[1], $this->sourceString);
			$this->sourceString .= ' ' . implode(' ', $match[1]);
		}
		preg_match_all('/alt="([^\"]*)"/si', $this->sourceString, $match);
		if( count($match[1]) > 0 ) {
			$this->sourceString .= ' ' . implode(' ', $match[1]);
		}
		// 移除 html 標籤
		$this->sourceString = strip_tags($this->sourceString);
	}

	private function StringClear() {
		// NCR 轉換回字符
		$this->sourceString = preg_replace_callback('/&#([0-9]{1,6}|x[0-9A-F]{1,5});/i', array(__CLASS__, 'decode_ncr_callback'), $this->sourceString);
		// html實體字符轉換回字符
		$this->sourceString = htmlspecialchars_decode($this->sourceString, ENT_QUOTES);
		// 移除特殊字符
		$sb_match	= "\r\n\t\x00\x01\x02\x03\x04\x05\x06\x07\x08\x0B\x0C\x0E\x0F\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1A\x1B\x1C\x1D\x1E\x1F\xC0\xC1\xF5\xF6\xF7\xF8\xF9\xFA\xFB\xFC\xFD\xFE\xFF";
		$sb_replace	= '                                              ';
		$this->sourceString = strtr($this->sourceString, $sb_match, $sb_replace);
	}

	private function decode_ncr_callback($matches) {
		$cp = (strncasecmp($matches[1], 'x', 1)) ? $matches[1] : hexdec(substr($matches[1], 1));
		if( $cp > 0xFFFF ) {
			return chr(0xF0 | ($cp >> 18)) . chr(0x80 | (($cp >> 12) & 0x3F)) . chr(0x80 | (($cp >> 6) & 0x3F)) . chr(0x80 | ($cp & 0x3F));
		} elseif( $cp > 0x7FF ) {
			return chr(0xE0 | ($cp >> 12)) . chr(0x80 | (($cp >> 6) & 0x3F)) . chr(0x80 | ($cp & 0x3F));
		} elseif( $cp > 0x7F ) {
			return chr(0xC0 | ($cp >> 6)) . chr(0x80 | ($cp & 0x3F));
		} else {
			return chr($cp);
		}
	}

	private function simpleString() {
		// 全型轉半型
		$sb_match = array('！', '＂', '＃', '＄', '％', '＆', '＇', '（', '）', '＊', '＋', '，', '－', '．', '／', '０', '１', '２', '３', '４', '５', '６', '７', '８', '９', '：', '；', '＜', '＝', '＞', '？', '＠', 'Ａ', 'Ｂ', 'Ｃ', 'Ｄ', 'Ｅ', 'Ｆ', 'Ｇ', 'Ｈ', 'Ｉ', 'Ｊ', 'Ｋ', 'Ｌ', 'Ｍ', 'Ｎ', 'Ｏ', 'Ｐ', 'Ｑ', 'Ｒ', 'Ｓ', 'Ｔ', 'Ｕ', 'Ｖ', 'Ｗ', 'Ｘ', 'Ｙ', 'Ｚ', '［', '＼', '］', '＾', '＿', '｀', 'ａ', 'ｂ', 'ｃ', 'ｄ', 'ｅ', 'ｆ', 'ｇ', 'ｈ', 'ｉ', 'ｊ', 'ｋ', 'ｌ', 'ｍ', 'ｎ', 'ｏ', 'ｐ', 'ｑ', 'ｒ', 'ｓ', 'ｔ', 'ｕ', 'ｖ', 'ｗ', 'ｘ', 'ｙ', 'ｚ', '｛', '｜', '｝', '～');
		$sb_replace = array('!', '"', '#', '$', '%', '&', '\'', '(', ')', '*', '+', ',', '-', '.', '/', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', ':', ';', '<', '=', '>', '?', '@', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '[', '\\', ']', '^', '_', '`', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '{', '|', '}', '~');
		$this->sourceString = str_replace($sb_match, $sb_replace, $this->sourceString);
		// 英文標點符號 轉 空白
		$sb_match = '!"#$%&\'()*+,/:;<=>?@[\\]^`{|}~';
		$sb_replace = '                           ';
		$this->sourceString = strtr($this->sourceString, $sb_match, $sb_replace);
		$this->sourceString = preg_replace_callback('/([^ ]*)([_.-])([^ ]*)/i', array(__CLASS__, 'word_cat_callback'), $this->sourceString);
		// 中文標點符號 轉 空白 
		$sb_match = array('。', '，', '‧', '？', '！', '：', '；', '、', '〈', '〉', '《', '》', '【', '】', '﹝', '﹞', '〝', '〞', '‘', '’', '“', '”', '『', '』', '「', '」');
		$sb_replace = array(' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ');
		$this->sourceString = str_replace($sb_match, $sb_replace, $this->sourceString);
		$this->sourceString = str_replace('  ', ' ', $this->sourceString);
		$this->simpleResult = array_filter(explode(' ', $this->sourceString));
		$this->simpleResult = array_count_values($this->simpleResult);
	}

	private function word_cat_callback($matches) {
		if( !preg_match('/[0-9a-z]/i', $matches[1]) || !preg_match('/[0-9a-z]/i', $matches[3]) ) {
			$matches[2] = ' ';
		}
		return $matches[1] . $matches[2] . $matches[3];
	}

	private function forward_n_gram() {
		foreach( $this->simpleResult as $string => $times ) {
			$words = array();
			$len = mb_strlen($string);
			$is_eng = false;
			for( $i = 0; $i < $len; $i++ ) {
				$word = mb_substr($string, $i, 1);
				while( preg_match('/[0-9a-z_\.\-]/i', $word) ) {
					if( is_array($words) ) {
						$this->forward_n_gram_make_words($words);
						$words = '';
					}
					$words .= $word;
					$i += 1;
					if( $i >= $len ) {
						break;
					}
					$word = mb_substr($string, $i, 1);
				}
				if( is_string($words) ) {
					if( strlen($words) > 1 ) {
						$this->n_gramResult[] = $words;
					}
					$words = array();
				} else {
					$words[] = $word;
				}
			}
			if( count($words) > 0 ) {
				$this->forward_n_gram_make_words($words);
			}
		}
		$this->n_gramResult = array_count_values($this->n_gramResult);
	}

	private function reverse_n_gram() {
		foreach( $this->simpleResult as $string => $times ) {
			$words = array();
			$len = mb_strlen($string);
			for( $i = $len - 1; $i >= 0; $i-- ) {
				$word = mb_substr($string, $i, 1);
				while( preg_match('/[0-9a-z_\.\-]/i', $word) ) {
					if( is_array($words) ) {
						$this->reverse_n_gram_make_words($words);
						$words = '';
					}
					$words = $word . $words;
					$i -= 1;
					if( $i < 0 ) {
						break;
					}
					$word = mb_substr($string, $i, 1);
				}
				if( is_string($words) ) {
					if( strlen($words) > 1 ) {
						$this->n_gramResult[] = $words;
					}
					$words = array();
				} else {
					$words[] = $word;
				}
			}
			if( count($words) > 0 ) {
				$this->reverse_n_gram_make_words($words);
			}
		}
		$this->n_gramResult = array_count_values($this->n_gramResult);
	}

	private function forward_n_gram_make_words($words) {
		$len = count($words) - 1;
		$word = implode('', $words);
		for( $i = 0; $i < $len; $i++ ) {
			for( $j = mb_strlen($word); $j > 1; $j-- ) {
				if( $j > $this->max_words ) {
					continue;
				}
				$temp = mb_substr($word, 0, $j);
				$this->n_gramResult[] = $temp;
				if( isset($this->ditc[$temp]) ) {
					$j--;
					$word = mb_substr($word, $j);
					break;
				}
			}
			$word = mb_substr($word, 1);
		}
	}

	private function reverse_n_gram_make_words($words) {
		$len = count($words) - 1;
		$word = implode('', array_reverse($words));
		for( $i = 0; $i < $len; $i++ ) {
			for( $j = mb_strlen($word); $j > 1; $j-- ) {
				if( $j > $this->max_words ) {
					continue;
				}
				$temp = mb_substr($word, 0 - $j);
				$this->n_gramResult[] = $temp;
				if( isset($this->ditc[$temp]) ) {
					$j--;
					$word = mb_substr($word, 0, 0 - $j);
					break;
				}
			}
			$word = mb_substr($word, 0, -1);
		}
	}

	private function makefinally() {
		$this->finallyResult = $this->n_gramResult;
		$ditc_sum = array_sum($this->ditc);
		$ngram_sum = array_sum($this->n_gramResult);
		$simple_sum = array_sum($this->simpleResult);
		foreach( $this->finallyResult as $key => $value ) {
			$this->finallyResult[$key] = $value / $ngram_sum;
			$word_weight = 0;
			$len = mb_strlen($key);
			for( $i = 0; $i < $len; $i++ ) {
				$word_weight += $words[mb_substr($key, $i, 1)];
			}
			if( isset($this->simpleResult[$key]) && $this->n_gramResult[$key] > 1 ) {
				$this->finallyResult[$key] += $this->simpleResult[$key] / $simple_sum;
			}
			if( isset($this->ditc[$key]) ) {
				$this->finallyResult[$key] += $this->ditc[$key] / $ditc_sum;
			}
		}
		$cat = array_values($this->finallyResult);
		arsort($this->finallyResult);
	}
}
?>